﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace Shooter {

	enum Kind {
		/// <summary>他と衝突システムで衝突しない背景の物体や機能</summary>
		Background,
		/// <summary>敵</summary>
		Receiver1,
		/// <summary>味方（自機）</summary>
		Reciever0,
		/// <summary>味方の弾</summary>
		Bullet0,
		/// <summary>敵の弾</summary>
		Bullet1,
		/// <summary>他と衝突システムで衝突しない前景の物体や機能</summary>
		Particle,
	}
	enum BlendMode {
		Opacity,
		Add,
	}

	/// <summary>
	/// 出来事
	/// </summary>
	interface Hap {
		void Apply(Actor actor);
	}

	static partial class Helper {
		public static void Apply(this List<Hap> haps, Actor actor) {
			foreach (var hap in haps) {
				hap.Apply(actor);
			}
		}
		public static void Clear(this StringBuilder text) {
			text.Remove(0, text.Length);
		}
	}

	class Mobile : Hap {
		public Vector2 Velocity;
		public virtual void Apply(Actor actor) {
			actor.Velocity = Velocity;
		}
	}

	abstract class Bullet : Hap {
		public abstract void Apply(Actor actor);
	}

	class DefaultBullet0 : Bullet {
		public override void Apply(Actor actor) {
			var item = actor.World.CreateActor();
			item.Blend = BlendMode.Add;
			item.Kind = Kind.Bullet0;
			item.Owner = actor;
			item.Clip = actor.World.Clips[1];
			item.Position = actor.Position + new Vector2(0x00, 0x00);
			item.Velocity = new Vector2(0, -10);
			item.Attack = 1;
			item.Life = 1;
			item.AutoDisposing = true;
			item.CollisionScale = 0.5f;
			actor.World.Add(item);
		}
	}
	class DefaultBullet1 : Bullet {
		public override void Apply(Actor actor) {
			var item = actor.World.CreateActor();
			item.Blend = BlendMode.Add;
			item.Kind = Kind.Bullet1;
			item.Target = actor.Target;
			item.Owner = actor;
			item.Clip = actor.World.Clips[3];
			item.Position = actor.Position + new Vector2(0x00, 0x00);
			item.Velocity = new Vector2(0, 4);
			item.Attack = 1;
			item.Life = 1;
			item.AutoDisposing = true;
			item.CollisionScale = 0.5f;
			Aim.Instance.Apply(item);
			actor.World.Add(item);
		}
	}

	class Aim : Hap {
		public static readonly Aim Instance = new Aim();
		protected Aim() { }
		public virtual void Apply(Actor actor) {
			if (actor.Target != null) {
				var direction = actor.Target.Position - actor.Position;
				if (direction.LengthSquared() != 0) {
					direction.Normalize();
					actor.Velocity = direction * actor.Velocity.Length();
				}
			}
		}
	}

	class Periodical : Hap {
		public Hap Launch;
		public int Span = 60;
		public int Time = 0;
		public virtual void Apply(Actor actor) {
			if (++Time >= Span) {
				Time = 0;
				Launch.Apply(actor);
			}
		}
	}
	class UserTrigger : Hap {
		public Hap Launch;
		public virtual void Apply(Actor actor) {
			if (actor.World.Input.IsKeyPress(Keys.X)) {
				Launch.Apply(actor);
			}
		}
	}

	class UserFollower : Mobile {
		public float Speed = 4;
		public override void Apply(Actor actor) {
			Velocity.X = 0;
			Velocity.Y = 0;
			if (actor.World.Input.CurrentKeyboardState.IsKeyDown(Keys.A)
				|| actor.World.Input.CurrentKeyboardState.IsKeyDown(Keys.Left)) {
				Velocity.X -= Speed;
			}
			if (actor.World.Input.CurrentKeyboardState.IsKeyDown(Keys.D)
				|| actor.World.Input.CurrentKeyboardState.IsKeyDown(Keys.Right)) {
				Velocity.X += Speed;
			}
			if (actor.World.Input.CurrentKeyboardState.IsKeyDown(Keys.W)
				|| actor.World.Input.CurrentKeyboardState.IsKeyDown(Keys.Up)) {
				Velocity.Y -= Speed;
			}
			if (actor.World.Input.CurrentKeyboardState.IsKeyDown(Keys.S)
				|| actor.World.Input.CurrentKeyboardState.IsKeyDown(Keys.Down)) {
				Velocity.Y += Speed;
			}
			base.Apply(actor);
		}
	}

	/// <summary>
	/// 画像の断片
	/// </summary>
	struct Clip {
		public Clip(Texture2D texture, Rectangle rectangle) {
			Texture = texture;
			Rectangle = rectangle;
		}
		public Texture2D Texture;
		public Rectangle Rectangle;
	}

	/// <summary>
	/// Clipの描画
	/// </summary>
	class ClipRenderer {
		public readonly SpriteBatch SpriteBatch;
		public readonly SpriteBatch AddBlendSpriteBatch;
		public ClipRenderer(GraphicsDevice device) {
			SpriteBatch = new SpriteBatch(device);
			AddBlendSpriteBatch = new SpriteBatch(device);
		}
		public void Draw(Clip clip, Vector2 position, float scale, BlendMode blend) {
			if (clip.Texture != null) {
				Vector2 origin = new Vector2(clip.Rectangle.Width * 0.5f, clip.Rectangle.Height * 0.5f);
				if (blend == BlendMode.Opacity) {
					SpriteBatch.Draw(clip.Texture, position, clip.Rectangle, Color.White,
						0, origin, scale, SpriteEffects.None, 0);
				} else {
					AddBlendSpriteBatch.Draw(clip.Texture, position, clip.Rectangle, Color.White,
						0, origin, scale, SpriteEffects.None, 0);
				}
			}
		}

		public void Begin() {
			SpriteBatch.Begin();
			AddBlendSpriteBatch.Begin(SpriteBlendMode.Additive);
		}
		public void End() {
			SpriteBatch.End();
			AddBlendSpriteBatch.End();
		}
	}

	/// <summary>
	/// メニューとスコア
	/// </summary>
	class GameMenu : Actor {
		readonly StringBuilder text = new StringBuilder();
		public readonly ScoreRanking Ranking = new ScoreRanking();
		public GameMenu(World world)
			: base(world) {
			Life = 1;
			Kind = Kind.Particle;
			Haps.Add(new PushXToStart());
			Haps.Add(new PushEscapeToExit());
		}
		public override void Draw(ClipRenderer renderer) {
			text.Clear();
			text.Append("GAME OVER\nPush (X) to Start.");
			var position = new Vector2(0, 300) + (new Vector2(800, 300) - World.Font.MeasureString(text)) * 0.5f;
			renderer.SpriteBatch.DrawString(World.Font, text, position, Color.White);
			Ranking.Draw(World, renderer);
		}
	}

	/// <summary>
	/// 入力
	/// </summary>
	class Input {
		public KeyboardState CurrentKeyboardState;
		public KeyboardState PreviousKeyboardState;

		public void Update() {
			PreviousKeyboardState = CurrentKeyboardState;
			CurrentKeyboardState = Keyboard.GetState();
		}
		public bool IsKeyPress(Keys key) {
			return CurrentKeyboardState.IsKeyDown(key) && PreviousKeyboardState.IsKeyUp(key);
		}
		public bool IsKeyRelease(Keys key) {
			return CurrentKeyboardState.IsKeyUp(key) && PreviousKeyboardState.IsKeyDown(key);
		}
	}

	class PushXToStart : Hap {
		public void Apply(Actor actor) {
			if (actor.World.Input.IsKeyPress(Keys.X)) {
				actor.World.Start();
				new GameStart().Apply(actor);
			}
		}
	}
	class PushEscapeToExit : Hap {
		public void Apply(Actor actor) {
			if (actor.World.Input.IsKeyPress(Keys.Escape)) {
				actor.World.Exit();
			}
		}
	}
	class EscapeToMenu : Hap {
		public bool Enabled;
		public void Apply(Actor actor) {
			if (Enabled) {
				if (actor.World.Input.IsKeyPress(Keys.Escape)) {
					GameOver.Instance.Apply(actor);
					actor.World.Add(actor.World.Menu);
				}
			}
		}
	}

	class Actor {
		internal void Initialize() {
			IsDisposed = false;
			AutoDisposing = false;
			TotalScore = 0;
			Score = 0;
			Owner = null;
			Target = null;
			worldRectangle = Rectangle.Empty;
			Position = Vector2.Zero;
			Velocity = Vector2.Zero;
			Life = 1;
			MaximumLife = 1;
			HealOverTime = 0;
			Attack = 0;
			Screenable = true;
			ScaleByLife = 0;
			ScaleBias = 1;
			Haps.Clear();
			DyingHaps.Clear();
			Kind = Kind.Background;
			Blend = BlendMode.Opacity;
			Clip.Rectangle = Rectangle.Empty;
			Clip.Texture = null;
			CollisionScale = 1.0f;
		}
		public bool IsDisposed { get; protected set; }
		public bool AutoDisposing;
		public int TotalScore;
		public int Score;
		public Actor Owner;
		public Actor Target;

		Rectangle worldRectangle;
		public Rectangle WorldRectangle {
			get { return worldRectangle; }
		}

		public readonly World World;
		public Vector2 Position;
		public Vector2 Velocity;
		public int Life;
		public int MaximumLife;
		public int HealOverTime;
		public int Attack;
		public bool Screenable;
		public float ScaleByLife;
		public float ScaleBias;

		public readonly List<Hap> Haps = new List<Hap>();
		public readonly List<Hap> DyingHaps = new List<Hap>();
		public Kind Kind;
		public BlendMode Blend;
		public Clip Clip;

		public float CollisionScale;

		public override string ToString() {
			return Kind.ToString();
		}

		internal Actor(World world) { World = world; Initialize(); }
		public virtual void Draw(ClipRenderer renderer) {
			renderer.Draw(Clip, Position, ScaleBias + ScaleByLife * Life, Blend);
		}

		public static int Clamp(int value, int minValue, int maxValue) {
			return Math.Min(Math.Max(value, minValue), maxValue);
		}
		public virtual void Update() {
			Haps.Apply(this);
			Position += Velocity;

			Life = Clamp(Life + HealOverTime, 0, MaximumLife);
			if (Life <= 0) {
				World.Remove(this);
				DyingHaps.Apply(this);
			}

			worldRectangle = Clip.Rectangle;
			worldRectangle.X = (int)Position.X;
			worldRectangle.Y = (int)Position.Y;

			var origin = worldRectangle.Center;

			worldRectangle = new Rectangle(
				worldRectangle.Left + (int)((worldRectangle.Left - origin.X) * CollisionScale),
				worldRectangle.Top + (int)((worldRectangle.Top - origin.Y) * CollisionScale),
				(int)(worldRectangle.Width * CollisionScale),
				(int)(worldRectangle.Height * CollisionScale));

			if (Screenable) {
				if (!World.ScreenOutRectangle.Intersects(worldRectangle)) {
					World.Remove(this);
				}
			}
		}

		internal void Dispose() {
			IsDisposed = true;
		}
	}

	/// <summary>
	/// レベル再生
	/// </summary>
	class LevelPlayer : Actor {
		public Level Level;

		public LevelPlayer(World world) : base(world) { Kind = Kind.Particle; }
		public int Time;
		public override void Update() {
			var haps = Level[Time];
			if (haps != null) {
				foreach (var hap in haps) {
					hap.Apply(this);
				}
			}
			Time++;
		}
	}

	/// <summary>
	/// レベル（ステージ）のタイムライン（敵の出現など）
	/// </summary>
	class Level {
		readonly Dictionary<int, List<Hap>> Timeline = new Dictionary<int, List<Hap>>();
		public void Add(int time, Hap hap) {
			List<Hap> haps;
			if (!Timeline.TryGetValue(time, out haps)) Timeline[time] = haps = new List<Hap>();
			haps.Add(hap);
		}
		public List<Hap> this[int time] {
			get {
				List<Hap> haps;
				Timeline.TryGetValue(time, out haps);
				return haps;
			}
			set {
				Timeline[time] = value;
			}
		}
	}

	class PopMob : Hap {
		public void Apply(Actor actor) {
			var item = actor.World.CreateActor();
			item.Kind = Kind.Receiver1;
			item.Clip = actor.World.Clips[2];
			item.Position = new Vector2(actor.World.Random.Next(0, 800), actor.World.Random.Next(-300, -300));
			item.Velocity = new Vector2(0, 2);
			item.Attack = 1;
			item.Life = 1;
			item.Score = 100;
			item.AutoDisposing = true;
			item.Haps.Add(new Periodical { Launch = new DefaultBullet1(), });
			item.Target = actor.World.Player;
			actor.World.Add(item);
		}
	}

	/// <summary>
	/// 指定のアイテムを出現させる。
	/// </summary>
	class Pop : Hap {
		public Pop(Actor item) { Item = item; }
		public readonly Actor Item;
		public void Apply(Actor actor) {
			actor.World.Add(Item);
		}
	}

	class PopStar : Hap {
		public void Apply(Actor actor) {
			var item = actor.World.CreateActor();
			item.Kind = Kind.Background;
			item.Blend = BlendMode.Add;
			item.Clip = actor.World.Clips[5];
			item.Position = new Vector2(actor.World.Random.Next(0, 800), -300);
			item.Velocity = new Vector2(0, 3 + (float)actor.World.Random.NextDouble() * 3);
			item.AutoDisposing = true;
			actor.World.Add(item);
		}
	}

	class StarNest : Actor {
		readonly Periodical Periodical = new Periodical { Launch = new PopStar(), Span = 2, };
		public StarNest(World world) : base(world) { }
		public override void Update() {
			Periodical.Apply(this);
			base.Update();
		}
	}

	class GameOver : Hap {
		public static readonly GameOver Instance = new GameOver();
		public void Apply(Actor actor) {
			actor.World.IsInGame = false;
			actor.World.Menu.Ranking.Join(new Score { Name = "YOU", Points = actor.TotalScore, });
			actor.World.Remove(actor);
			actor.World.Remove(actor.World.ScoreDisplay);
		}
	}
	class GameStart : Hap {
		public void Apply(Actor actor) {
			actor.World.IsInGame = true;
			actor.World.Add(actor.World.ScoreDisplay);
		}
	}

	class World {
		public bool IsInGame;
		public event EventHandler Exiting;
		Rectangle screenOutRectangle = new Rectangle(-800 / 2, -600 / 2, 800 + 800, 600 + 600);
		Rectangle ringRectangle = new Rectangle(80, 60, 800 - 80 * 2, 600 - 60 * 2);
		public Rectangle ScreenOutRectangle {
			get { return screenOutRectangle; }
		}
		public Rectangle RingRectangle {
			get { return ringRectangle; }
		}
		public List<Actor> this[Kind kind] {
			get { return itemGroups[(int)kind]; }
		}
		public readonly Input Input = new Input();
		readonly GraphicsDevice GraphicsDevice;
		public readonly SpriteFont Font;
		public readonly SpriteFont SmallFont;

		public Actor Player;
		public readonly List<Clip> Clips = new List<Clip>();

		readonly ClipRenderer renderer;

		public World(GraphicsDevice device, ContentManager content) {
			this.GraphicsDevice = device;
			Font = content.Load<SpriteFont>("DefaultFont");
			SmallFont = content.Load<SpriteFont>("SmallFont");

			var texture = content.Load<Texture2D>("actor");
			Clips.Add(new Clip(texture, new Rectangle(0x000, 0x000, 0x80, 0x80))); // player
			Clips.Add(new Clip(texture, new Rectangle(0x080, 0x000, 0x20, 0x50))); // bullet 0
			Clips.Add(new Clip(texture, new Rectangle(0x000, 0x080, 0x80, 0x80))); // mob
			Clips.Add(new Clip(texture, new Rectangle(0x080, 0x080, 0x20, 0x50))); // bullet 1
			Clips.Add(new Clip(texture, new Rectangle(0x000, 0x100, 0x80, 0x80))); // explosion
			Clips.Add(new Clip(texture, new Rectangle(0x000, 0x180, 0x20, 0x20))); // star
			Clips.Add(new Clip(texture, new Rectangle(0x020, 0x180, 0x20, 0x20))); // shard
			LevelPlayer = new LevelPlayer(this);
			LevelPlayer.Level = new Level();
			LevelPlayer.Level.Add(0, new Pop(new StarNest(this)));
			for (int i = 0; i < 10; i++) {
				LevelPlayer.Level.Add(100 + i * 100, new PopMob());
			}
			renderer = new ClipRenderer(device);

			Menu = new GameMenu(this);
			ScoreDisplay = new ScoreDisplay(this);

			Start();
			Add(Menu);
		}

		public readonly GameMenu Menu;
		public readonly ScoreDisplay ScoreDisplay;

		public void Start() {
			Clear();
			LevelPlayer.Time = 0;
			var player = CreateActor();
			player.Kind = Kind.Reciever0;
			player.Clip = Clips[0];
			player.Life = 1;
			player.Position = new Vector2(0x100, 0x200);
			player.AutoDisposing = true;
			player.CollisionScale = 0.25f;
			player.Haps.Add(new UserFollower());
			player.Haps.Add(new UserTrigger() { Launch = new DefaultBullet0(), });
			player.Haps.Add(new EscapeToMenu() { Enabled = true, });
			player.DyingHaps.Add(new Pop(Menu));
			player.DyingHaps.Add(GameOver.Instance);
			Add(player);
			Player = player;
			Add(LevelPlayer);
		}

		public readonly LevelPlayer LevelPlayer;

		public void Clear() {
			foreach (var group in itemGroups) {
				foreach (var item in group) {
					Remove(item);
				}
			}
		}

		public readonly Random Random = new Random();

		void Collides(List<Actor> defenses, List<Actor> offenses) {
			for (int j = 0; j < offenses.Count; j++) {
				var o = offenses[j];
				for (int i = 0; i < defenses.Count; i++) {
					var d = defenses[i];
					if (d.WorldRectangle.Intersects(o.WorldRectangle)) {
						d.Life -= o.Attack;
						o.Life -= o.Attack;
						if (d.Life <= 0) {
							d.Life = 0;
							d.DyingHaps.Apply(d);
							Remove(d);
							Explode(d.Position);
							if (o.World.IsInGame) o.Owner.TotalScore += d.Score;
						}
						if (o.Life <= 0) {
							o.Life = 0;
							Remove(o);
							Explode(o.Position);
						}
					}
				}
			}
		}

		void Explode(Vector2 position) {
			{
				var item = CreateActor();
				item.Kind = Kind.Particle;
				item.Blend = BlendMode.Add;
				item.Position = position;
				item.Clip = Clips[4];
				item.Life = 20;
				item.MaximumLife = 20;
				item.HealOverTime = -1;
				item.ScaleBias = 0;
				item.ScaleByLife = 1.0f / 20;
				item.AutoDisposing = true;
				Add(item);
			}

			for (int i = 0; i < 30; i++) {
				var item = CreateActor();
				item.Kind = Kind.Particle;
				item.Blend = BlendMode.Add;
				item.Position = position;
				item.Velocity = new Vector2((float)Random.NextDouble() * 2 - 1, (float)Random.NextDouble() * 2 - 1) * 5;
				item.Clip = Clips[6];
				item.Life = 30;
				item.MaximumLife = 30;
				item.HealOverTime = -1;
				item.ScaleBias = 0;
				item.ScaleByLife = 1.0f / 30;
				item.AutoDisposing = true;
				Add(item);
			}
		}

		readonly Stack<Actor> itemPool = new Stack<Actor>(1024);
		public Actor CreateActor() {
			Actor item;
			if (itemPool.Count <= 0) item = new Actor(this);
			else item = itemPool.Pop();
			item.Initialize();
			return item;
		}
		public void DisposeActor(Actor item) {
			item.Dispose();
			itemPool.Push(item);
		}

		public void Draw() {
			foreach (var group in itemGroups) {
				foreach (var item in group) {
					renderer.Begin();
					item.Draw(renderer);
					renderer.End();
				}
			}
		}

		public void Update() {
			Input.Update();
			foreach (var group in itemGroups) {
				foreach (var item in group) {
					if (item.IsDisposed) {
						throw new ObjectDisposedException(item.ToString());
					}
					item.Update();
				}
			}

			Collides(this[Kind.Reciever0], this[Kind.Bullet1]);
			Collides(this[Kind.Receiver1], this[Kind.Bullet0]);

			foreach (var item in removingItems) {
				foreach (var group in itemGroups) {
					if (group.Remove(item) && item.AutoDisposing) {
						DisposeActor(item);
					}
				}
			}
			removingItems.Clear();
			foreach (var item in addingItems) {
				if (!itemGroups[(int)item.Kind].Contains(item))
					itemGroups[(int)item.Kind].Add(item);
			}
			addingItems.Clear();
		}

		readonly List<Actor>[] itemGroups = new List<Actor>[]{
			new List<Actor>(),
			new List<Actor>(),
			new List<Actor>(),
			new List<Actor>(),
			new List<Actor>(),
			new List<Actor>(),
		};

		readonly List<Actor> addingItems = new List<Actor>();
		readonly List<Actor> removingItems = new List<Actor>();

		public void Add(Actor item) {
			if (item != null)
				addingItems.Add(item);
		}
		public void Remove(Actor item) {
			if (item != null)
				removingItems.Add(item);
		}

		public void Exit() {
			if (Exiting != null) Exiting(this, EventArgs.Empty);
		}
	}

}
